import logging as log
from typing import cast, Any, Callable, Type, TypeVar

import olca_schema as o
import requests

from .protocol import E, IpcProtocol, IpcResult

T = TypeVar("T")


class RestClient(IpcProtocol):
    def __init__(self, endpoint: str, **kwargs):
        self.endpoint = endpoint if endpoint.endswith("/") else endpoint + "/"
        self.req_args = kwargs

    def _get(self, path, transform: Callable[[Any], T]) -> T | None:
        resp = requests.get(self.endpoint + path, **self.req_args)
        if _not_ok(resp):
            log.error("ERROR: GET %s failed: %s", path, resp.text)
            return None
        return transform(resp.json())

    def _post(
        self, path, transform: Callable[[Any], T], data: Any | None = None
    ) -> T | None:
        resp = requests.post(self.endpoint + path, json=data, **self.req_args)
        if _not_ok(resp):
            log.error("ERROR: POST %s failed: %s", path, resp.text)
            return None
        return transform(resp.json())

    def _get_each(self, path, transform: Callable[[Any], T]) -> list[T]:
        resp = requests.get(self.endpoint + path, **self.req_args)
        if _not_ok(resp):
            log.error("ERROR: GET %s failed: %s", path, resp.text)
            return []
        return [transform(x) for x in resp.json()]

    def get(
        self,
        model_type: Type[E],
        uid: str | None = None,
        name: str | None = None,
    ) -> E | None:
        if not uid and not name:
            log.error("no ID or name given")
            return None
        if uid:
            path = f"data/{_path_of(model_type)}/{uid}"
        else:
            path = f"data/{_path_of(model_type)}/name/{name}"
        return cast(E, self._get(path, model_type.from_dict))

    def get_all(self, model_type: Type[E]) -> list[E]:
        xs = self._get_each(
            f"data/{_path_of(model_type)}/all",
            model_type.from_dict,
        )
        return cast(list[E], xs)

    def get_descriptors(self, model_type: Type[E]) -> list[o.Ref]:
        return self._get_each(
            f"data/{_path_of(model_type)}",
            o.Ref.from_dict,
        )

    def get_descriptor(
        self,
        model_type: Type[E],
        uid: str | None = None,
        name: str | None = None,
    ) -> o.Ref | None:
        if uid is None and name is None:
            log.error("error: no uuid or name given")
            return None
        if uid is not None:
            return self._get(
                f"data/{_path_of(model_type)}/{uid}/info", o.Ref.from_dict
            )
        if name is not None:
            return self._get(
                f"data/{_path_of(model_type)}/name/{name}/info",
                o.Ref.from_dict,
            )

    def get_providers(
        self, flow: o.Ref | o.Flow | None = None
    ) -> list[o.TechFlow]:
        if flow is not None:
            path = f"data/providers/{flow.id}"
        else:
            path = "data/providers"
        return self._get_each(path, o.TechFlow.from_dict)

    def get_parameters(
        self, model_type: Type[E], uid: str
    ) -> list[o.Parameter] | list[o.ParameterRedef]:
        path = f"data/{_path_of(model_type)}/{uid}/parameters"
        if model_type in (o.ProductSystem, o.Project):
            return self._get_each(path, o.ParameterRedef.from_dict)
        else:
            return self._get_each(path, o.Parameter.from_dict)

    def put(self, model: o.RootEntity) -> o.Ref | None:
        resp = requests.put(
            f"{self.endpoint}data/{_path_of(model.__class__)}",
            json=model.to_dict(),
            **self.req_args,
        )
        if _not_ok(resp):
            log.error("failed to upload entity: %s", resp.text)
            return None
        return o.Ref.from_dict(resp.json())

    def create_product_system(
        self,
        process: o.Ref | o.Process,
        config: o.LinkingConfig | None = None,
    ) -> o.Ref | None:
        params: dict[str, Any] = {"process": o.as_ref(process).to_dict()}
        if config is not None:
            params["config"] = config.to_dict()
        return self._post("data/create-system", o.Ref.from_dict, params)

    def delete(self, model: o.RootEntity | o.Ref) -> o.Ref | None:
        if isinstance(model, o.Ref):
            t = model.model_type
        else:
            t = model.__class__.__name__
        path = ""
        for char in t:
            if char.isupper() and len(path) > 0:
                path += "-"
            path += char
        url = f"{self.endpoint}data/{path.lower()}/{model.id}"
        resp = requests.delete(url, **self.req_args)
        if _not_ok(resp):
            log.error("failed to delete model: %s", resp.text)
            return None
        return o.Ref.from_dict(resp.json())

    def calculate(self, setup: o.CalculationSetup) -> IpcResult | None:
        state = self._post(
            "result/calculate", o.ResultState.from_dict, setup.to_dict()
        )
        if state is None:
            return None
        return Result(self, state)

    def simulate(self, setup: o.CalculationSetup) -> IpcResult | None:
        state = self._post(
            "result/simulate", o.ResultState.from_dict, setup.to_dict()
        )
        if state is None:
            return None
        return Result(self, state)


class Result(IpcResult):
    def __init__(self, client: RestClient, state: o.ResultState):
        self.client = client
        self.uid = state.id
        self.error: o.ResultState | None = None
        if state.error:
            self.error = state

    def get_state(self) -> o.ResultState:
        if self.error is not None:
            return self.error
        state = self._get("state", o.ResultState.from_dict)
        if state is None:
            self.error = o.ResultState(
                id=self.uid,
                error="no result state could be retreived from server",
            )
            return self.error
        if state.error:
            self.error = state
        return state

    def simulate_next(self) -> o.ResultState:
        if self.error is not None:
            return self.error
        state = self.client._post(
            f"result/{self.uid}/simulate/next", o.ResultState.from_dict
        )
        if state is None:
            self.err = o.ResultState(
                id=self.uid, error="failed to run simulation"
            )
            return self.err
        return state

    def dispose(self) -> o.ResultState:
        if self.error:
            return self.error
        state = self.client._post(
            f"result/{self.uid}/dispose", o.ResultState.from_dict
        )
        if state is None:
            self.error = o.ResultState(
                id=self.uid, error="dispose did not return state"
            )
            return self.error
        return state

    def get_demand(self) -> o.TechFlowValue | None:
        return self._get("demand", o.TechFlowValue.from_dict)

    def get_tech_flows(self) -> list[o.TechFlow]:
        return self._get_each("tech-flows", o.TechFlow.from_dict)

    def get_envi_flows(self) -> list[o.EnviFlow]:
        return self._get_each("envi-flows", o.EnviFlow.from_dict)

    def get_impact_categories(self) -> list[o.Ref]:
        return self._get_each("impact-categories", o.Ref.from_dict)

    # region: tech-flows

    def get_total_requirements(self) -> list[o.TechFlowValue]:
        return self._get_each("total-requirements", o.TechFlowValue.from_dict)

    def get_total_requirements_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        return self._get_each(
            f"total-requirements-of/{_tech_id(tech_flow)}",
            o.TechFlowValue.from_dict,
        )

    def get_scaling_factors(self) -> list[o.TechFlowValue]:
        return self._get_each("scaling-factors", o.TechFlowValue.from_dict)

    def get_scaled_tech_flows_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        return self._get_each(
            f"scaled-tech-flows-of/{_tech_id(tech_flow)}",
            o.TechFlowValue.from_dict,
        )

    def get_unscaled_tech_flows_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        return self._get_each(
            f"unscaled-tech-flows-of/{_tech_id(tech_flow)}",
            o.TechFlowValue.from_dict,
        )

    # endregion

    def get_total_flows(self) -> list[o.EnviFlowValue]:
        return self._get_each("total-flows", o.EnviFlowValue.from_dict)

    def get_total_flow_value_of(
        self, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue | None:
        return self._get(
            f"total-flow-value-of/{_envi_id(envi_flow)}",
            o.EnviFlowValue.from_dict,
        )

    def get_flow_contributions_of(
        self, envi_flow: o.EnviFlow
    ) -> list[o.TechFlowValue]:
        return self._get_each(
            f"flow-contributions-of/{_envi_id(envi_flow)}",
            o.TechFlowValue.from_dict,
        )

    def get_direct_interventions_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        return self._get_each(
            f"direct-interventions-of/{_tech_id(tech_flow)}",
            o.EnviFlowValue.from_dict,
        )

    def get_direct_intervention_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        val = self._get(
            f"direct-intervention-of/{_envi_id(envi_flow)}/{_tech_id(tech_flow)}",
            o.EnviFlowValue.from_dict,
        )
        if val is None:
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return val

    def get_flow_intensities_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlow]:
        return self._get_each(
            f"flow-intensities-of/{_tech_id(tech_flow)}", o.EnviFlow.from_dict
        )

    def get_flow_intensity_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        val = self._get(
            f"flow-intensity-of/{_envi_id(envi_flow)}/{_tech_id(tech_flow)}",
            o.EnviFlowValue.from_dict,
        )
        if val is None:
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return val

    def get_total_interventions_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        return self._get_each(
            f"total-interventions-of/{_tech_id(tech_flow)}",
            o.EnviFlowValue.from_dict,
        )

    def get_total_intervention_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        val = self._get(
            f"total-intervention-of/{_envi_id(envi_flow)}/{_tech_id(tech_flow)}",
            o.EnviFlowValue.from_dict,
        )
        if val is None:
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return val

    def get_total_impacts(self) -> list[o.ImpactValue]:
        return self._get_each(f"total-impacts", o.ImpactValue.from_dict)

    def get_total_impact_value_of(
        self, impact_category: o.Ref
    ) -> o.ImpactValue:
        val = self._get(
            f"total-impact-value-of/{impact_category.id}",
            o.ImpactValue.from_dict,
        )
        if val is None:
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return val

    def get_normalized_impacts(self) -> list[o.ImpactValue]:
        return self._get_each(f"normalized-impacts", o.ImpactValue.from_dict)

    def get_weighted_impacts(self) -> list[o.ImpactValue]:
        return self._get_each(f"weighted-impacts", o.ImpactValue.from_dict)

    def get_impact_contributions_of(
        self, impact_category: o.Ref
    ) -> list[o.TechFlowValue]:
        return self._get_each(
            f"impact-contributions-of/{impact_category.id}",
            o.TechFlowValue.from_dict,
        )

    def get_direct_impacts_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        return self._get_each(
            f"direct-impacts-of/{_tech_id(tech_flow)}",
            o.ImpactValue.from_dict,
        )

    def get_direct_impact_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        val = self._get(
            f"direct-impact-of/{impact_category.id}/{_tech_id(tech_flow)}",
            o.ImpactValue.from_dict,
        )
        if val is None:
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return val

    def get_impact_intensities_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        return self._get_each(
            f"impact-intensities-of/{_tech_id(tech_flow)}",
            o.ImpactValue.from_dict,
        )

    def get_impact_intensity_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        val = self._get(
            f"impact-intensity-of/{impact_category.id}/{_tech_id(tech_flow)}",
            o.ImpactValue.from_dict,
        )
        if val is None:
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return val

    def get_total_impacts_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        return self._get_each(
            f"total-impacts-of/{_tech_id(tech_flow)}", o.ImpactValue.from_dict
        )

    def get_total_impact_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        val = self._get(
            f"total-impact-of/{impact_category.id}/{_tech_id(tech_flow)}",
            o.ImpactValue.from_dict,
        )
        if val is None:
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return val

    def get_impact_factors_of(
        self, impact_category: o.Ref
    ) -> list[o.EnviFlowValue]:
        return self._get_each(
            f"impact-factors-of/{impact_category.id}",
            o.EnviFlowValue.from_dict,
        )

    def get_impact_factor_of(
        self, impact_category: o.Ref, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue:
        val = self._get(
            f"impact-factor-of/{impact_category.id}/{_envi_id(envi_flow)}",
            o.EnviFlowValue.from_dict,
        )
        if val is None:
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return val

    def get_flow_impacts_of(
        self, impact_category: o.Ref
    ) -> list[o.EnviFlowValue]:
        return self._get_each(
            f"flow-impacts-of/{impact_category.id}", o.EnviFlowValue.from_dict
        )

    def get_flow_impact_of(
        self, impact_category: o.Ref, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue:
        val = self._get(
            f"flow-impact-of/{impact_category.id}/{_envi_id(envi_flow)}",
            o.EnviFlowValue.from_dict,
        )
        if val is None:
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return val

    def get_total_costs(self) -> o.CostValue:
        val = self._get(f"total-costs", o.CostValue.from_dict)
        if val is None:
            return o.CostValue(amount=0)
        return val

    def get_cost_contributions(self) -> list[o.TechFlowValue]:
        return self._get_each(f"cost-contributions", o.TechFlowValue.from_dict)

    def get_direct_costs_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        val = self._get(
            f"direct-costs-of/{_tech_id(tech_flow)}", o.CostValue.from_dict
        )
        if val is None:
            return o.CostValue(amount=0)
        return val

    def get_cost_intensities_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        val = self._get(
            f"cost-intensities-of/{_tech_id(tech_flow)}",
            o.CostValue.from_dict,
        )
        if val is None:
            return o.CostValue(amount=0)
        return val

    def get_total_costs_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        val = self._get(
            f"total-costs-of/{_tech_id(tech_flow)}", o.CostValue.from_dict
        )
        if val is None:
            return o.CostValue(amount=0)
        return val

    def _get(self, path: str, transform: Callable[[Any], T]) -> T | None:
        return self.client._get(f"result/{self.uid}/{path}", transform)

    def _get_each(self, path: str, transform: Callable[[Any], T]) -> list[T]:
        return self.client._get_each(f"result/{self.uid}/{path}", transform)


def _not_ok(resp: requests.Response) -> bool:
    if resp.status_code == 200:
        return False
    log.error("response status != 200; message=%s", resp.text)
    return True


def _tech_id(tech_flow: o.TechFlow) -> str:
    tech_id = ""
    if tech_flow.provider and tech_flow.provider.id:
        tech_id = tech_flow.provider.id
    if tech_flow.flow and tech_flow.flow.id:
        tech_id += "::" + tech_flow.flow.id
    return tech_id


def _envi_id(envi_flow: o.EnviFlow) -> str:
    envi_id = ""
    if envi_flow.flow and envi_flow.flow.id:
        envi_id = envi_flow.flow.id
    if envi_flow.location and envi_flow.location.id:
        envi_id += "::" + envi_flow.location.id
    return envi_id


def _path_of(model_type: Type[E]) -> str | None:
    if model_type is None:
        log.error("no model type given")
        return None
    match model_type:
        case o.Actor:
            return "actors"
        case o.Currency:
            return "currencies"
        case o.DQSystem:
            return "dq-systems"
        case o.Epd:
            return "epds"
        case o.Flow:
            return "flows"
        case o.FlowProperty:
            return "flow-properties"
        case o.ImpactCategory:
            return "impact-categories"
        case o.ImpactMethod:
            return "impact-methods"
        case o.Location:
            return "locations"
        case o.Parameter:
            return "parameters"
        case o.Process:
            return "processes"
        case o.ProductSystem:
            return "product-systems"
        case o.Project:
            return "projects"
        case o.Result:
            return "results"
        case o.SocialIndicator:
            return "social-indicators"
        case o.Source:
            return "sources"
        case o.UnitGroup:
            return "unit-groups"
        case _:
            log.error("unknown root entity type: %s", model_type)
            return None



import unittest
import statistics as stats

import olca_schema as o
from config import client


class MCSimTest(unittest.TestCase):
    def test_sim(self):

        # create units and flows
        units = o.new_unit_group("Units of mass", "kg")
        mass = o.new_flow_property("Mass", units)
        p = o.new_product("p", mass)
        q = o.new_product("q", mass)
        e = o.new_elementary_flow("e", mass)

        # create the process
        process = o.new_process("P")
        qref = o.new_output(process, p, amount=0.5)
        qref.is_quantitative_reference = True
        o.new_output(process, q, amount=0.5)
        o.new_output(process, e, amount=2).uncertainty = o.Uncertainty(
            distribution_type=o.UncertaintyType.NORMAL_DISTRIBUTION,
            mean=2,
            sd=0.1,
        )
        o.new_physical_allocation_factor(process, p, 0.2)
        o.new_physical_allocation_factor(process, q, 0.8)

        # create the LCIA indicator and method
        i = o.new_impact_category("i")
        o.new_impact_factor(i, e, value=2)
        method = o.new_impact_method("M", i)

        client.put_all(units, mass, p, q, e, process, i, method)

        setup = o.CalculationSetup(
            allocation=o.AllocationType.PHYSICAL_ALLOCATION,
            target=o.as_ref(process),
            amount=1,
            impact_method=o.as_ref(method),
        )

        # run simulations and collect indicator results
        xs = []
        result = client.simulate(setup)

        def put_next():
            result.wait_until_ready()
            xi = result.get_total_impact_value_of(o.as_ref(i))
            xs.append(xi.amount)

        put_next()
        for _ in range(0, 9):
            result.simulate_next()
            put_next()

        mean = stats.mean(xs)
        self.assertAlmostEqual(1.6, mean, delta=0.4)
        # print(xs)

        # clean up
        result.dispose()
        client.delete_all(method, i, process, e, p, q, mass, units)


if __name__ == "__main__":
    unittest.main()



import olca_ipc as ipc
import olca_ipc.protocol as protocol
import olca_ipc.rest as rest

# configure here the client against which
# the test suite should be executed
client: protocol.IpcProtocol
use_rest = False
if use_rest:
    client = rest.RestClient("http://localhost:8080")
else:
    client = ipc.Client(8080)



import unittest

import olca_schema as o

from config import client


class ParamTest(unittest.TestCase):
    def test_global_params(self):
        x = o.new_parameter("X", 4.0)
        y = o.new_parameter("Y", "X * 10")
        client.put(x)
        client.put(y)

        x = client.get(o.Parameter, x.id)
        y = client.get(o.Parameter, y.id)
        self.assertEqual("X", x.name)
        self.assertEqual(o.ParameterScope.GLOBAL_SCOPE, x.parameter_scope)
        self.assertEqual(4, x.value)
        self.assertTrue(x.is_input_parameter)
        client.delete(x)

        self.assertEqual("Y", y.name)
        self.assertEqual("X * 10", y.formula)
        self.assertEqual(o.ParameterScope.GLOBAL_SCOPE, y.parameter_scope)
        self.assertFalse(y.is_input_parameter)
        client.delete(y)


if __name__ == "__main__":
    unittest.main()



import unittest

import olca_schema as o
import olca_ipc.upstream_tree as u

from config import client


class TestUpstreamTree(unittest.TestCase):
    def test_upstream_tree(self):
        units = o.new_unit_group("Units of mass", "kg")
        mass = o.new_flow_property("Mass", units)
        e = o.new_elementary_flow("e", mass)
        p = o.new_product("p", mass)
        q = o.new_product("q", mass)

        P = o.new_process("P")
        Q = o.new_process("Q")
        o.new_output(P, p, amount=1).is_quantitative_reference = True
        o.new_input(P, q, amount=1).default_provider = Q.to_ref()
        o.new_output(P, e, amount=2)
        o.new_output(Q, q, amount=1).is_quantitative_reference = True
        o.new_input(Q, p, amount=0.5).default_provider = P.to_ref()
        o.new_output(Q, e, amount=4)

        client.put_all(units, mass, e, p, q, P, Q)

        setup = o.CalculationSetup(
            target=o.Ref(ref_type=o.RefType.Process, id=P.id)
        )
        result = client.calculate(setup)
        result.wait_until_ready()

        envi_flow = next(
            filter(lambda ei: ei.flow.id == e.id, result.get_envi_flows())
        )

        root = u.UpstreamTree(result, envi_flow=envi_flow).root
        self.assertAlmostEqual(12, root.total)
        self.assertAlmostEqual(2, root.direct)

        l1 = root.childs[0]
        self.assertAlmostEqual(10, l1.total)
        self.assertAlmostEqual(4, l1.direct)

        l2 = l1.childs[0]
        self.assertAlmostEqual(6, l2.total)
        self.assertAlmostEqual(1, l2.direct)

        l3 = l2.childs[0]
        self.assertAlmostEqual(5, l3.total)
        self.assertAlmostEqual(2, l3.direct)

        l4 = l3.childs[0]
        self.assertAlmostEqual(3, l4.total)
        self.assertAlmostEqual(0.5, l4.direct)

        l5 = l4.childs[0]
        self.assertAlmostEqual(2.5, l5.total)
        self.assertAlmostEqual(1.0, l5.direct)

        client.delete_all(Q, P, q, p, e, mass, units)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass

import olca_schema as o
from .protocol import IpcResult


_vec = dict[str, float]


class UpstreamTree:
    def __init__(
        self,
        result: IpcResult,
        envi_flow: o.EnviFlow | None = None,
        impact_category: o.Ref | None = None,
        is_for_costs: bool = False,
    ):
        self.result = result

        if envi_flow is not None:
            self._envi_flow = envi_flow
            cons = result.get_flow_contributions_of(envi_flow)
        elif impact_category is not None:
            self._impact_category = impact_category
            cons = result.get_impact_contributions_of(impact_category)
        elif is_for_costs:
            self._is_for_costs = is_for_costs
            cons = result.get_cost_contributions()
        else:
            raise AssertionError(
                "an intervention flow, impact "
                + "category, cost flag must be provided"
            )

        (self._diag, self._direct) = _unscaled_vecs(result, cons)
        d = result.get_demand()
        self.root = self._node_of(d.tech_flow, d.amount)

    def _node_of(self, provider: o.TechFlow, required_amount: float):
        key = _key(provider)
        diag_j = self._diag.get(key, 0.0)
        if required_amount == 0 or diag_j == 0:
            return UpstreamNode(
                tree=self,
                provider=provider,
                required_amount=0,
                total=0,
                direct=0,
                scaling=0,
            )
        ij = self._intensity_of(provider)
        sj = required_amount / diag_j
        direct = sj * self._direct.get(key, 0)
        return UpstreamNode(
            tree=self,
            provider=provider,
            required_amount=required_amount,
            total=ij * required_amount,
            direct=direct,
            scaling=sj,
        )

    def _intensity_of(self, provider: o.TechFlow) -> float:
        def _v(
            v: o.EnviFlowValue | o.ImpactValue | o.CostValue | None,
        ) -> float:
            if v is None or v.amount is None:
                return 0
            return v.amount

        if self._envi_flow is not None:
            return _v(
                self.result.get_flow_intensity_of(self._envi_flow, provider)
            )
        if self._impact_category is not None:
            return _v(
                self.result.get_impact_intensity_of(
                    self._impact_category, provider
                )
            )
        if self._is_for_costs:
            return _v(self.result.get_cost_intensities_of(provider))
        return 0


@dataclass
class UpstreamNode:
    tree: UpstreamTree
    provider: o.TechFlow
    scaling: float
    required_amount: float
    total: float
    direct: float
    _childs: list["UpstreamNode"] | None = None

    @property
    def childs(self) -> list["UpstreamNode"]:
        if self._childs is not None:
            return self._childs
        childs = []
        for ai in self.tree.result.get_unscaled_tech_flows_of(self.provider):
            if (
                ai.tech_flow is None
                or _eq(ai.tech_flow, self.provider)
                or ai.amount is None
                or ai.amount == 0
            ):
                continue
            val = -self.scaling * ai.amount
            childs.append(self.tree._node_of(ai.tech_flow, val))
        self._childs = childs
        return childs


def _unscaled_vecs(
    result: IpcResult, cons: list[o.TechFlowValue]
) -> tuple[_vec, _vec]:
    s: _vec = {}
    t: _vec = {}
    d: _vec = {}
    for sf in result.get_scaling_factors():
        s[_key(sf.tech_flow)] = sf.amount or 0
    for ti in result.get_total_requirements():
        t[_key(ti.tech_flow)] = ti.amount or 0
    for ci in cons:
        d[_key(ci.tech_flow)] = ci.amount or 0

    diag: _vec = {}
    direct: _vec = {}
    for (key, ti) in t.items():
        si = s.get(key, 0.0)
        di = d.get(key, 0.0)
        if si == 0 or ti == 0:
            diag[key] = 0
        else:
            diag[key] = ti / si
            direct[key] = di / si
    return (diag, direct)


def _key(tech_flow: o.TechFlow | None) -> str:
    if (
        tech_flow is None
        or tech_flow.provider is None
        or tech_flow.flow is None
    ):
        return ""
    return f"{tech_flow.provider.id}/{tech_flow.flow.id}"


def _eq(t1: o.TechFlow | None, t2: o.TechFlow | None) -> bool:
    return (
        t1 is not None
        and t2 is not None
        and t1.provider is not None
        and t2.provider is not None
        and t1.provider.id == t2.provider.id
        and t1.flow is not None
        and t2.flow is not None
        and t1.flow.id == t2.flow.id
    )



import time

from abc import abstractmethod
from typing import TypeVar, Type

import olca_schema as o

E = TypeVar("E", bound=o.RootEntity)


class IpcProtocol:
    @abstractmethod
    def get(
        self,
        model_type: Type[E],
        uid: str | None = None,
        name: str | None = None,
    ) -> E | None:
        pass

    @abstractmethod
    def get_all(self, model_type: Type[E]) -> list[E]:
        pass

    @abstractmethod
    def get_descriptors(self, model_type: Type[E]) -> list[o.Ref]:
        pass

    @abstractmethod
    def get_descriptor(
        self,
        model_type: Type[E],
        uid: str | None = None,
        name: str | None = None,
    ) -> o.Ref | None:
        pass

    def find(self, model_type: Type[E], name: str) -> o.Ref | None:
        for d in self.get_descriptors(model_type):
            if d.name == name:
                return d

    @abstractmethod
    def get_providers(
        self, flow: o.Ref | o.Flow | None = None
    ) -> list[o.TechFlow]:
        pass

    @abstractmethod
    def get_parameters(
        self, model_type: Type[E], uid: str
    ) -> list[o.Parameter | o.ParameterRedef]:
        pass

    @abstractmethod
    def put(self, model: o.RootEntity) -> o.Ref | None:
        pass

    def put_all(self, *models: o.RootEntity):
        for model in models:
            self.put(model)

    @abstractmethod
    def create_product_system(
        self,
        process: o.Ref | o.Process,
        config: o.LinkingConfig | None = None,
    ) -> o.Ref | None:
        pass

    @abstractmethod
    def delete(self, model: o.RootEntity | o.Ref) -> o.Ref | None:
        pass

    def delete_all(self, *models: o.RootEntity | o.Ref):
        for model in models:
            self.delete(model)

    @abstractmethod
    def calculate(self, setup: o.CalculationSetup) -> "IpcResult":
        pass

    @abstractmethod
    def simulate(self, setup: o.CalculationSetup) -> "IpcResult":
        pass


class IpcResult:
    @abstractmethod
    def get_state(self) -> o.ResultState:
        pass

    @abstractmethod
    def simulate_next(self) -> o.ResultState:
        pass

    def wait_until_ready(self) -> o.ResultState:
        state = self.get_state()
        if not state.is_scheduled:
            return state
        while state.is_scheduled:
            time.sleep(0.5)
            state = self.get_state()
            if not state.is_scheduled:
                return state
        return o.ResultState(error="did not finished")

    @abstractmethod
    def dispose(self):
        pass

    @abstractmethod
    def get_demand(self) -> o.TechFlowValue | None:
        pass

    @abstractmethod
    def get_tech_flows(self) -> list[o.TechFlow]:
        pass

    @abstractmethod
    def get_envi_flows(self) -> list[o.EnviFlow]:
        pass

    @abstractmethod
    def get_impact_categories(self) -> list[o.Ref]:
        pass

    # region: tech-flows

    @abstractmethod
    def get_total_requirements(self) -> list[o.TechFlowValue]:
        pass

    @abstractmethod
    def get_total_requirements_of(
        self, tech_flow: o.TechFlow
    ) -> o.TechFlowValue:
        pass

    @abstractmethod
    def get_scaling_factors(self) -> list[o.TechFlowValue]:
        pass

    @abstractmethod
    def get_scaled_tech_flows_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        pass

    @abstractmethod
    def get_unscaled_tech_flows_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        pass

    # endregion

    @abstractmethod
    def get_total_flows(self) -> list[o.EnviFlowValue]:
        pass

    @abstractmethod
    def get_total_flow_value_of(self, envi_flow: o.EnviFlow) -> o.EnviFlowValue:
        pass

    @abstractmethod
    def get_flow_contributions_of(
        self, envi_flow: o.EnviFlow
    ) -> list[o.TechFlowValue]:
        pass

    @abstractmethod
    def get_direct_interventions_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        pass

    @abstractmethod
    def get_direct_intervention_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        pass

    @abstractmethod
    def get_flow_intensities_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlow]:
        pass

    @abstractmethod
    def get_flow_intensity_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        pass

    @abstractmethod
    def get_total_interventions_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        pass

    @abstractmethod
    def get_total_intervention_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        pass

    # region: impacts

    @abstractmethod
    def get_total_impacts(self) -> list[o.ImpactValue]:
        pass

    @abstractmethod
    def get_total_impact_value_of(
        self, impact_category: o.Ref
    ) -> o.ImpactValue:
        pass

    @abstractmethod
    def get_normalized_impacts(self) -> list[o.ImpactValue]:
        pass

    @abstractmethod
    def get_weighted_impacts(self) -> list[o.ImpactValue]:
        pass

    @abstractmethod
    def get_impact_contributions_of(
        self, impact_category: o.Ref
    ) -> list[o.TechFlowValue]:
        pass

    @abstractmethod
    def get_direct_impacts_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        pass

    @abstractmethod
    def get_direct_impact_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        pass

    @abstractmethod
    def get_impact_intensities_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        pass

    @abstractmethod
    def get_impact_intensity_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        pass

    @abstractmethod
    def get_total_impacts_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        pass

    @abstractmethod
    def get_total_impact_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        pass

    @abstractmethod
    def get_impact_factors_of(
        self, impact_category: o.Ref
    ) -> list[o.EnviFlowValue]:
        pass

    @abstractmethod
    def get_impact_factor_of(
        self, impact_category: o.Ref, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue:
        pass

    @abstractmethod
    def get_flow_impacts_of(
        self, impact_category: o.Ref
    ) -> list[o.EnviFlowValue]:
        pass

    @abstractmethod
    def get_flow_impact_of(
        self, impact_category: o.Ref, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue:
        pass

    # endregion: impacts

    @abstractmethod
    def get_total_costs(self) -> o.CostValue:
        pass

    @abstractmethod
    def get_cost_contributions(self) -> list[o.TechFlowValue]:
        pass

    @abstractmethod
    def get_direct_costs_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        pass

    @abstractmethod
    def get_cost_intensities_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        pass

    @abstractmethod
    def get_total_costs_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        pass



import unittest

import olca_schema as o
import olca_schema.units as units

from config import client


class AllocationTest(unittest.TestCase):
    """
    Tests the calculation of results for different allocation methods.
    """

    def test_example(self):

        # get our quantity and unit, assuming that we are connected to an
        # openLCA database that contains these reference data
        mass = client.get(o.FlowProperty, name="Mass")
        assert mass is not None
        kg = units.unit_ref("kg")
        assert kg is not None

        # create the flows p, q, e
        p = o.new_product("p", mass)
        q = o.new_product("q", mass)
        e = o.new_elementary_flow("e", mass)
        for f in [p, q, e]:
            client.put(f)

        # create the process with inputs and outputs
        process = o.new_process("P")
        assert process.id is not None
        o.new_output(process, p, 1.0, kg).is_quantitative_reference = True
        o.new_output(process, q, 1.0, kg)
        ex = o.new_output(process, e, 1.0, kg)

        # define the allocation factors
        process.default_allocation_method = o.AllocationType.CAUSAL_ALLOCATION
        o.new_physical_allocation_factor(process, p, 0.25)
        o.new_physical_allocation_factor(process, q, 0.75)
        o.new_economic_allocation_factor(process, p, 0.4)
        o.new_economic_allocation_factor(process, q, 0.6)
        o.new_causal_allocation_factor(process, p, 0.1, ex)
        o.new_causal_allocation_factor(process, q, 0.9, ex)

        # save the process and create the product system
        client.put(process)
        system = client.create_product_system(process)
        assert system is not None
        setup = o.CalculationSetup()
        setup.target = o.Ref(ref_type=o.RefType.ProductSystem, id=system.id)
        setup.amount = 500
        setup.unit = units.unit_ref("g")

        expected = [
            (o.AllocationType.PHYSICAL_ALLOCATION, 0.25 / 2),
            (o.AllocationType.ECONOMIC_ALLOCATION, 0.4 / 2),
            (o.AllocationType.CAUSAL_ALLOCATION, 0.1 / 2),
            (o.AllocationType.USE_DEFAULT_ALLOCATION, 0.1 / 2),
            (o.AllocationType.NO_ALLOCATION, 0.5),
            (None, 0.5),
        ]
        for (method, value) in expected:
            print("test with allocation method = %s ..." % method)
            setup.allocation = method
            result = client.calculate(setup)
            result.wait_until_ready()
            # get the result for 'e'; when there is no allocation applied
            # we also have a result for ' q' in our list
            fr = next(
                r
                for r in result.get_total_flows()
                if r.envi_flow.flow.id == e.id
            )
            self.assertFalse(fr.envi_flow.is_input)
            self.assertAlmostEqual(value, fr.amount)
            result.dispose()
            print(" ... success")

        # set drop to False if you want to inspect the process in openLCA
        drop = False
        if drop:
            for each in (system, process, p, q, e):
                client.delete(each)


if __name__ == "__main__":
    unittest.main()



import unittest

import olca_schema as o

from config import client


class SimpleCalculationTest(unittest.TestCase):
    def setUp(self):
        units = o.new_unit_group("Units of mass", "kg")
        mass = o.new_flow_property("Mass", units)
        e = o.new_elementary_flow("e", mass)
        p = o.new_product("p", mass)
        process = o.new_process("P")
        o.new_output(process, p, amount=1).is_quantitative_reference = True
        o.new_output(process, e, amount=21)

        i = o.new_impact_category("i")
        o.new_impact_factor(i, e, 0.5)
        method = o.new_impact_method("M", i)

        self.process = process
        self.e = e
        self.entities = [units, mass, e, p, process, i, method]
        for entity in self.entities:
            client.put(entity)

        setup = o.CalculationSetup(
            target=process.to_ref(),
            impact_method=method.to_ref(),
            amount=2,  # kg
        )
        self.result = client.calculate(setup)
        self.result.wait_until_ready()

    def tearDown(self):
        self.result.dispose()
        self.entities.reverse()
        for entity in self.entities:
            client.delete(entity)

    def test_demand(self):
        demand = self.result.get_demand()
        self.assertEqual("P", demand.tech_flow.provider.name)
        self.assertEqual("p", demand.tech_flow.flow.name)
        self.assertEqual(2, demand.amount)

    def test_tech_flow(self):
        tech_flow = self.get_tech_flow()
        self.assertEqual("P", tech_flow.provider.name)
        self.assertEqual("p", tech_flow.flow.name)
        self.assertEqual("kg", tech_flow.flow.ref_unit)

    def test_envi_flow(self):
        envi_flow = self.get_envi_flow()
        self.assertEqual("e", envi_flow.flow.name)
        self.assertEqual("kg", envi_flow.flow.ref_unit)

    def test_impact(self):
        impact = self.get_impact()
        self.assertEqual("i", impact.name)

    def test_tech_flow_values(self):
        r = self.result
        seq = [
            # todo: scaling factors etc.
            (r.get_total_requirements(), 2)
        ]
        for (values, expected) in seq:
            value: o.TechFlowValue = next(
                filter(
                    lambda x: x.tech_flow.provider.id == self.process.id, values
                )
            )
            self.assertEqual(expected, value.amount)

    def test_envi_flow_values(self):
        tech_flow = self.get_tech_flow()
        xs = [
            self.result.get_total_flows(),
            self.result.get_direct_interventions_of(tech_flow),
            self.result.get_total_interventions_of(tech_flow),
        ]
        for r in xs:
            value: o.EnviFlowValue = next(
                filter(lambda x: x.envi_flow.flow.id == self.e.id, r)
            )
            self.assertEqual(42, value.amount)

    def test_envi_flow_value(self):
        envi_flow = self.get_envi_flow()
        tech_flow = self.get_tech_flow()
        values = [
            self.result.get_total_flow_value_of(envi_flow),
            self.result.get_total_intervention_of(envi_flow, tech_flow),
            self.result.get_direct_intervention_of(envi_flow, tech_flow),
        ]
        for v in values:
            self.assertEqual(42, v.amount)
            self.assertEqual("e", v.envi_flow.flow.name)
            self.assertEqual("kg", v.envi_flow.flow.ref_unit)

    def test_impact_values(self):
        tech_flow = self.get_tech_flow()
        xs = [
            self.result.get_total_impacts(),
            self.result.get_direct_impacts_of(tech_flow),
            self.result.get_total_impacts_of(tech_flow),
        ]
        for r in xs:
            value: o.ImpactValue = r[0]
            self.assertEqual(21, value.amount)

    def test_impact_value(self):
        tech_flow = self.get_tech_flow()
        impact = self.get_impact()
        values = [
            self.result.get_total_impact_value_of(impact),
            self.result.get_direct_impact_of(impact, tech_flow),
            self.result.get_total_impact_of(impact, tech_flow),
        ]
        for v in values:
            self.assertEqual(21, v.amount)
            self.assertEqual("i", v.impact_category.name)

    def get_tech_flow(self) -> o.TechFlow:
        tech_flows = self.result.get_tech_flows()
        for tech_flow in tech_flows:
            if tech_flow.provider and tech_flow.provider.id == self.process.id:
                return tech_flow
        raise ValueError("tech. flow not found")

    def get_envi_flow(self) -> o.EnviFlow:
        envi_flows = self.result.get_envi_flows()
        for envi_flow in envi_flows:
            if envi_flow.flow and envi_flow.flow.id == self.e.id:
                return envi_flow
        raise ValueError("envi. flow not found")

    def get_impact(self) -> o.Ref:
        impacts = self.result.get_impact_categories()
        for i in impacts:
            if i.name == "i":
                return i
        raise ValueError("impact category not found")


if __name__ == "__main__":
    unittest.main()



import unittest

import olca_schema as o

from config import client


@unittest.SkipTest
class GetTest(unittest.TestCase):
    """A test for getting entities from the server.

    Only works when a server is running with a database that contains the
    standard flow properties and units.
    """

    def test_get_for_name(self):
        mass = client.get(o.FlowProperty, name="Mass")
        self.assertEqual("Mass", mass.name)

    def test_get_for_id(self):
        mass = client.get(o.FlowProperty, name="Mass")
        group = client.get(o.UnitGroup, mass.unit_group.id)
        self.assertEqual(
            "kg", next(filter(lambda u: u.is_ref_unit, group.units)).name
        )

    def test_get_all(self):
        props = client.get_all(o.FlowProperty)
        mass = next(filter(lambda p: p.name == "Mass", props))
        self.assertIsNotNone(mass)


if __name__ == "__main__":
    unittest.main()



import unittest

import olca_schema as o

from config import client


class TestQRef(unittest.TestCase):
    def test_qref(self):
        mass = client.get(o.FlowProperty, name="Mass")
        steel = o.new_product("Steel", mass)
        process = o.new_process("Steel production")
        o.new_exchange(process, steel).is_quantitative_reference = True
        client.put_all(steel, process)

        process = client.get(o.Process, process.id)
        qref = next(e for e in process.exchanges if e.is_quantitative_reference)
        self.assertEqual(steel.id, qref.flow.id)
        self.assertEqual(1, qref.amount)

        client.delete_all(process, steel)


if __name__ == "__main__":
    unittest.main()



import logging as log
from dataclasses import dataclass
from typing import cast, Any, Callable, Optional, Tuple, Type, TypeVar

import requests
import olca_schema as o

from .protocol import E, IpcProtocol, IpcResult

_T = TypeVar("_T")


class Client(IpcProtocol):
    """
    A client to communicate with an openLCA IPC server over the JSON-RPC
    protocol.
    """

    def __init__(self, port: int = 8080):
        self.url = "http://localhost:%i" % port
        self.next_id = 1

    def get(
        self,
        model_type: Type[E],
        uid: str | None = None,
        name: str | None = None,
    ) -> E | None:
        if not uid and not name:
            log.error("no ID or name given")
            return None
        params = {"@type": model_type.__name__}
        if uid is not None:
            params["@id"] = uid
        if name is not None:
            params["name"] = name
        result, err = self.rpc_call("data/get", params)
        if err:
            log.warning("failed to get entity of type %s: %s", model_type, err)
            return None
        return cast(E, model_type.from_dict(result))

    def get_all(self, model_type: Type[E]) -> list[E]:
        params = {"@type": model_type.__name__}
        result, err = self.rpc_call("data/get/all", params)
        if err:
            log.error("failed to get all of type %s: %s", model_type, err)
        return cast(list[E], [model_type.from_dict(r) for r in result])

    def get_descriptors(self, model_type: Type[E]) -> list[o.Ref]:
        params = {"@type": model_type.__name__}
        result, err = self.rpc_call("data/get/descriptors", params)
        if err:
            log.error(
                "failed to get descriptors of type %s: %s", model_type, err
            )
            return []
        return [o.Ref.from_dict(r) for r in result]

    def get_descriptor(
        self,
        model_type: Type[E],
        uid: str | None = None,
        name: str | None = None,
    ) -> o.Ref | None:
        params = {"@type": model_type.__name__}
        if uid is not None:
            params["@id"] = uid
        if name is not None:
            params["name"] = name
        result, err = self.rpc_call("data/get/descriptor", params)
        if err:
            log.error("failed to get descriptor: %s", err)
            return None
        return o.Ref.from_dict(result)

    def get_providers(
        self, flow: o.Ref | o.Flow | None = None
    ) -> list[o.TechFlow]:
        if flow is None:
            params = {}
        else:
            params = {
                "@type": "Flow",
                "@id": flow.id,
                "name": flow.name,
            }
        providers, err = self.rpc_call("data/get/providers", params)
        if err:
            log.error("failed to get providers: %s", err)
            return []
        return [o.TechFlow.from_dict(d) for d in providers]

    def get_parameters(
        self, model_type: Type[E], uid: str
    ) -> list[o.Parameter | o.ParameterRedef]:
        params = {"@type": model_type.__name__, "@id": uid}
        params, err = self.rpc_call("data/get/parameters", params)
        if err:
            log.error("failed to get parameters of %s id=%d", model_type, uid)
            return []
        if model_type in (o.Process, o.ImpactCategory):
            return [o.Parameter.from_dict(p) for p in params]
        else:
            return [o.ParameterRedef.from_dict(p) for p in params]

    def put(self, model: o.RootEntity) -> o.Ref | None:
        if model is None:
            return None
        resp, err = self.rpc_call("data/put", model.to_dict())
        if err:
            log.error("failed to insert model: %s", err)
            return None
        return o.Ref.from_dict(resp)

    def create_product_system(
        self,
        process: o.Ref | o.Process,
        config: o.LinkingConfig | None = None,
    ) -> o.Ref | None:
        conf = config
        if conf is None:
            conf = o.LinkingConfig(
                prefer_unit_processes=False,
                provider_linking=o.ProviderLinking.PREFER_DEFAULTS,
            )
        r, err = self.rpc_call(
            "data/create/system",
            {
                "process": o.as_ref(process).to_dict(),
                "config": conf.to_dict(),
            },
        )
        if err:
            log.error("failed to create product system: %s", err)
            return None
        return o.Ref.from_dict(r)

    def delete(self, model: o.RootEntity | o.Ref) -> o.Ref | None:
        if model is None:
            return
        ref = o.as_ref(model).to_dict()
        resp, err = self.rpc_call("data/delete", ref)
        if err:
            log.error("failed to delete model: %s", err)
            return None
        return o.Ref.from_dict(resp)

    def calculate(self, setup: o.CalculationSetup) -> "Result":
        resp, err = self.rpc_call("result/calculate", setup.to_dict())
        if err:
            return Result(
                uid="", client=self, error=o.ResultState(id="", error=err)
            )
        state = o.ResultState.from_dict(resp)
        return Result(uid=state.id, client=self, error=None)

    def simulate(self, setup: o.CalculationSetup) -> "Result":
        resp, err = self.rpc_call("result/simulate", setup.to_dict())
        if err:
            return Result(
                uid="", client=self, error=o.ResultState(id="", error=err)
            )
        state = o.ResultState.from_dict(resp)
        return Result(uid=state.id, client=self, error=None)

    def rpc_call(
        self, method: str, params: Any = None
    ) -> Tuple[Any, Optional[str]]:
        """
        Performs a JSON-RPC request with the given parameters.

        It returns a tuple (result, error).
        """
        req = {
            "jsonrpc": "2.0",
            "id": self.next_id,
            "method": method,
        }
        if params is not None:
            req["params"] = params
        self.next_id += 1
        resp: dict = requests.post(self.url, json=req).json()
        err: dict | None = resp.get("error")
        if err is not None:
            err_msg = "%i: %s" % (err.get("code"), err.get("message"))
            return None, err_msg
        result = resp.get("result")
        if result is None:
            return None, "No error and no result: invalid JSON-RPC response"
        return result, None

    def _call(
        self, method: str, transform: Callable[[Any], _T], data: Any = None
    ) -> _T | None:
        (resp, err) = self.rpc_call(method, data)
        if err is not None:
            log.error("failed to call method %s: %s", method, err)
            return None
        return transform(resp)

    def _call_each(
        self, method: str, transform: Callable[[Any], _T], data: Any = None
    ) -> list[_T]:
        (resp, err) = self.rpc_call(method, data)
        if err is not None:
            log.error("failed to call method %s: %s", method, err)
            return []
        return [transform(r) for r in resp]


@dataclass
class Result(IpcResult):
    uid: str
    client: "Client"
    error: Optional[o.ResultState]

    def get_state(self) -> o.ResultState:
        if self.error is not None:
            return self.error
        (state, err) = self.client.rpc_call("result/state", {"@id": self.uid})
        if err:
            self.err = o.ResultState(id=self.uid, error=err)
            return self.err
        return o.ResultState.from_dict(state)

    def simulate_next(self) -> o.ResultState:
        if self.error is not None:
            return self.error
        (state, err) = self.client.rpc_call(
            "result/simulate/next", {"@id": self.uid}
        )
        if err:
            self.err = o.ResultState(id=self.uid, error=err)
            return self.err
        return o.ResultState.from_dict(state)

    def dispose(self):
        if self.error is not None:
            return
        self.client.rpc_call("result/dispose", {"@id": self.uid})

    def get_demand(self) -> o.TechFlowValue | None:
        (data, err) = self.client.rpc_call("result/demand", {"@id": self.uid})
        if err:
            log.error("failed to get demand: %s", err)
            return None
        return o.TechFlowValue.from_dict(data)

    def get_tech_flows(self) -> list[o.TechFlow]:
        (data, err) = self.client.rpc_call(
            "result/tech-flows", {"@id": self.uid}
        )
        if err:
            log.error("failed to query /tech-flows: %s", err)
            return []
        return [o.TechFlow.from_dict(d) for d in data]

    def get_envi_flows(self) -> list[o.EnviFlow]:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/envi-flows", args)
        if err:
            log.error("request envi-flows failed: %s", err)
            return []
        return [o.EnviFlow.from_dict(d) for d in r]

    def get_impact_categories(self) -> list[o.Ref]:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/impact-categories", args)
        if err:
            log.error("request impact-categories failed: %s", err)
            return []
        return [o.Ref.from_dict(d) for d in r]

    # region: tech-flows

    def get_total_requirements(self) -> list[o.TechFlowValue]:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/total-requirements", args)
        if err:
            log.error("request total-requirements failed: %s", err)
            return []
        return [o.TechFlowValue.from_dict(d) for d in r]

    def get_total_requirements_of(
        self, tech_flow: o.TechFlow
    ) -> o.TechFlowValue:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/total-requirements-of", args)
        if err:
            log.error("request total-requirements-of failed: %s", err)
            return o.TechFlowValue(amount=0, tech_flow=tech_flow)
        return o.TechFlowValue.from_dict(r)

    def get_scaling_factors(self) -> list[o.TechFlowValue]:
        return self.client._call_each(
            "result/scaling-factors",
            o.TechFlowValue.from_dict,
            {"@id": self.uid},
        )

    def get_scaled_tech_flows_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        return self.client._call_each(
            "result/scaled-tech-flows-of",
            o.TechFlowValue.from_dict,
            {"@id": self.uid, "techFlow": tech_flow.to_dict()},
        )

    def get_unscaled_tech_flows_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.TechFlowValue]:
        return self.client._call_each(
            "result/unscaled-tech-flows-of",
            o.TechFlowValue.from_dict,
            {"@id": self.uid, "techFlow": tech_flow.to_dict()},
        )

    # endregion

    # region: flows

    def get_total_flows(self) -> list[o.EnviFlowValue]:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/total-flows", args)
        if err:
            log.error("request total-flows failed: %s", err)
            return []
        return [o.EnviFlowValue.from_dict(d) for d in r]

    def get_total_flow_value_of(self, envi_flow: o.EnviFlow) -> o.EnviFlowValue:
        args = {"@id": self.uid, "enviFlow": envi_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/total-flow-value-of", args)
        if err:
            log.error("request total-flow-value-of failed: %s", err)
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return o.EnviFlowValue.from_dict(r)

    def get_flow_contributions_of(
        self, envi_flow: o.EnviFlow
    ) -> list[o.TechFlowValue]:
        args = {"@id": self.uid, "enviFlow": envi_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/flow-contributions-of", args)
        if err:
            log.error("request direct-flow-values-of failed: %s", err)
            return []
        return [o.TechFlowValue.from_dict(d) for d in r]

    def get_direct_interventions_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/direct-interventions-of", args)
        if err:
            log.error("request direct-flows-of failed: %s", err)
            return []
        return [o.EnviFlowValue.from_dict(d) for d in r]

    def get_direct_intervention_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        args = {
            "@id": self.uid,
            "enviFlow": envi_flow.to_dict(),
            "techFlow": tech_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/direct-intervention-of", args)
        if err:
            log.error("request direct-flow-of failed: %s", err)
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return o.EnviFlowValue.from_dict(r)

    def get_flow_intensities_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/flow-intensities-of", args)
        if err:
            log.error("request total-flows-of-one failed: %s", err)
            return []
        return [o.EnviFlowValue.from_dict(d) for d in r]

    def get_flow_intensity_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        args = {
            "@id": self.uid,
            "enviFlow": envi_flow.to_dict(),
            "techFlow": tech_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/flow-intensity-of", args)
        if err:
            log.error("request total-flow-of-one failed: %s", err)
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return o.EnviFlowValue.from_dict(r)

    def get_total_interventions_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.EnviFlowValue]:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/total-interventions-of", args)
        if err:
            log.error("request total-flows-of failed: %s", err)
            return []
        return [o.EnviFlowValue.from_dict(d) for d in r]

    def get_total_intervention_of(
        self, envi_flow: o.EnviFlow, tech_flow: o.TechFlow
    ) -> o.EnviFlowValue:
        args = {
            "@id": self.uid,
            "enviFlow": envi_flow.to_dict(),
            "techFlow": tech_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/total-intervention-of", args)
        if err:
            log.error("request total-flow-of failed: %s", err)
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return o.EnviFlowValue.from_dict(r)

    # endregion: flows

    def get_total_impacts(self) -> list[o.ImpactValue]:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/total-impacts", args)
        if err:
            log.error("request total-impacts failed: %s", err)
            return []
        return [o.ImpactValue.from_dict(d) for d in r]

    def get_total_impact_value_of(
        self, impact_category: o.Ref
    ) -> o.ImpactValue:
        args = {"@id": self.uid, "impactCategory": impact_category.to_dict()}
        (r, err) = self.client.rpc_call("result/total-impact-value-of", args)
        if err:
            log.error("request total-impact-value-of failed: %s", err)
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return o.ImpactValue.from_dict(r)

    def get_normalized_impacts(self) -> list[o.ImpactValue]:
        (r, err) = self.client.rpc_call(
            "result/total-impacts/normalized", {"@id": self.uid}
        )
        if err:
            log.error("failed to get normalized impacts: %s", err)
            return []
        return [o.ImpactValue.from_dict(d) for d in r]

    def get_weighted_impacts(self) -> list[o.ImpactValue]:
        (r, err) = self.client.rpc_call(
            "result/total-impacts/weighted", {"@id": self.uid}
        )
        if err:
            log.error("failed to get weighted impacts: %s", err)
            return []
        return [o.ImpactValue.from_dict(d) for d in r]

    def get_impact_contributions_of(
        self, impact_category: o.Ref
    ) -> list[o.TechFlowValue]:
        args = {"@id": self.uid, "impactCategory": impact_category.to_dict()}
        (r, err) = self.client.rpc_call("result/impact-contributions-of", args)
        if err:
            log.error("request direct-impact-values-of failed: %s", err)
            return []
        return [o.TechFlowValue.from_dict(d) for d in r]

    def get_direct_impacts_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/direct-impacts-of", args)
        if err:
            log.error("request direct-impacts-of failed: %s", err)
            return []
        return [o.ImpactValue.from_dict(d) for d in r]

    def get_direct_impact_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        args = {
            "@id": self.uid,
            "impactCategory": impact_category.to_dict(),
            "techFlow": tech_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/direct-impact-of", args)
        if err:
            log.error("request direct-impact-of failed: %s", err)
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return o.ImpactValue.from_dict(r)

    def get_impact_intensities_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/total-impacts-of-one", args)
        if err:
            log.error("request total-impacts-of-one failed: %s", err)
            return []
        return [o.ImpactValue.from_dict(d) for d in r]

    def get_impact_intensity_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        args = {
            "@id": self.uid,
            "impactCategory": impact_category.to_dict(),
            "techFlow": tech_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/impact-intensity-of", args)
        if err:
            log.error("request total-impact-of-one failed: %s", err)
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return o.ImpactValue.from_dict(r)

    def get_total_impacts_of(
        self, tech_flow: o.TechFlow
    ) -> list[o.ImpactValue]:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/total-impacts-of", args)
        if err:
            log.error("request total-impacts-of failed: %s", err)
            return []
        return [o.ImpactValue.from_dict(d) for d in r]

    def get_total_impact_of(
        self, impact_category: o.Ref, tech_flow: o.TechFlow
    ) -> o.ImpactValue:
        args = {
            "@id": self.uid,
            "impactCategory": impact_category.to_dict(),
            "techFlow": tech_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/total-impact-of", args)
        if err:
            log.error("request total-impact-of failed: %s", err)
            return o.ImpactValue(amount=0, impact_category=impact_category)
        return o.ImpactValue.from_dict(r)

    def get_impact_factors_of(
        self, impact_category: o.Ref
    ) -> list[o.EnviFlowValue]:
        args = {"@id": self.uid, "impactCategory": impact_category.to_dict()}
        (r, err) = self.client.rpc_call("result/impact-factors-of", args)
        if err:
            log.error("request impact-factors-of failed: %s", err)
            return []
        return [o.EnviFlowValue.from_dict(d) for d in r]

    def get_impact_factor_of(
        self, impact_category: o.Ref, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue:
        args = {
            "@id": self.uid,
            "impactCategory": impact_category.to_dict(),
            "enviFlow": envi_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/impact-factor-of", args)
        if err:
            log.error("request impact-factor-of failed: %s", err)
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return o.EnviFlowValue.from_dict(r)

    def get_flow_impacts_of(
        self, impact_category: o.Ref
    ) -> list[o.EnviFlowValue]:
        args = {"@id": self.uid, "impactCategory": impact_category.to_dict()}
        (r, err) = self.client.rpc_call("result/flow-impacts-of", args)
        if err:
            log.error("request flow-impacts-of failed: %s", err)
            return []
        return [o.EnviFlowValue.from_dict(d) for d in r]

    def get_flow_impact_of(
        self, impact_category: o.Ref, envi_flow: o.EnviFlow
    ) -> o.EnviFlowValue:
        args = {
            "@id": self.uid,
            "impactCategory": impact_category.to_dict(),
            "enviFlow": envi_flow.to_dict(),
        }
        (r, err) = self.client.rpc_call("result/flow-impact-of", args)
        if err:
            log.error("request flow-impact-of failed: %s", err)
            return o.EnviFlowValue(amount=0, envi_flow=envi_flow)
        return o.EnviFlowValue.from_dict(r)

    def get_total_costs(self) -> o.CostValue:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/total-costs", args)
        if err:
            log.error("request total-costs failed: %s", err)
            return o.CostValue(amount=0)
        return o.CostValue.from_dict(r)

    def get_cost_contributions(self) -> list[o.TechFlowValue]:
        args = {"@id": self.uid}
        (r, err) = self.client.rpc_call("result/cost-contributions", args)
        if err:
            log.error("request direct-cost-values failed: %s", err)
            return []
        return [o.TechFlowValue.from_dict(d) for d in r]

    def get_direct_costs_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/direct-costs-of", args)
        if err:
            log.error("request direct-costs-of failed: %s", err)
            return o.CostValue(amount=0)
        return o.CostValue.from_dict(r)

    def get_cost_intensities_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/cost-intensities-of", args)
        if err:
            log.error("request total-costs-of-one failed: %s", err)
            return o.CostValue(amount=0)
        return o.CostValue.from_dict(r)

    def get_total_costs_of(self, tech_flow: o.TechFlow) -> o.CostValue:
        args = {"@id": self.uid, "techFlow": tech_flow.to_dict()}
        (r, err) = self.client.rpc_call("result/total-costs-of", args)
        if err:
            log.error("request total-costs-of failed: %s", err)
            return o.CostValue(amount=0)
        return o.CostValue.from_dict(r)



import unittest

import olca_schema as o

from config import client


class CrudTest(unittest.TestCase):
    def test_unit_group(self):

        # create a unit group
        units = o.new_unit_group("Units of mass", "kg")
        client.put(units)
        units = client.get(o.UnitGroup, units.id)
        self.assertEqual("kg", units.units[0].name)

        # create a flow property
        mass = o.new_flow_property("Mass", units)
        client.put(mass)
        mass = client.get(o.FlowProperty, mass.id)
        self.assertEqual(units.id, mass.unit_group.id)

        # update the unit group
        units.default_flow_property = mass.to_ref()
        client.put(units)
        units = client.get(o.UnitGroup, units.id)
        self.assertEqual(mass.id, units.default_flow_property.id)

        # delete everything
        client.delete(mass)
        client.delete(units)
        self.assertIsNone(client.get(o.UnitGroup, units.id))
        self.assertIsNone(client.get(o.FlowProperty, mass.id))

    def test_process(self):
        units = o.new_unit_group("Mass units", "kg")
        mass = o.new_flow_property("Mass", units)
        steel = o.new_product("Steel", mass)
        process = o.new_process("Steel production")
        o.new_output(process, steel).is_quantitative_reference = True

        for e in [units, mass, steel, process]:
            client.put(e)
        assert process.id is not None
        process = client.get(o.Process, process.id)

        output = process.exchanges[0]
        self.assertEqual(steel.id, output.flow.id)
        self.assertTrue(output.is_quantitative_reference)
        self.assertFalse(output.is_input)
        self.assertFalse(output.is_avoided_product)
        self.assertEqual(1.0, output.amount)

        for e in [process, steel, mass, units]:
            client.delete(e)


if __name__ == "__main__":
    unittest.main()
